#! /usr/bin/env python
from typing import Dict, List, Any, Generator
from datetime import datetime as dt
from collections import deque
import multiprocessing as mp
import logging
import signal
import time
import math
import json
import pdb
import sys

import plotext as plt
import docker

log = logging.getLogger(__file__)

shared_memory_manager = mp.Manager()

client = docker.from_env()
data_queue = shared_memory_manager.Queue()


def start_container_collection(name):
    log.debug(f"starting stats for {name}...")
    container = client.containers.get(name)
    try:
        for data in container.stats(decode=True):
            data['time'] = time.time()
            data_queue.put(data)
    except (KeyboardInterrupt, SystemExit) as err:
        ...
    log.debug(f"finished retrieving stats for {name}")


def process_record(record: Dict) -> Generator[Dict, None, None]:
    def get_block_io_mb_stats() -> Dict:
        ret_val = {}
        for stat in record['blkio_stats']['io_service_bytes_recursive']:
            op = stat['op'].lower()
            if op in {'read', 'write'}:
                ret_val[f"block_io_mb_{op}"] = stat['value'] * 1e-6
        return ret_val


    def get_cpu_stats() -> Dict:
        current_cpu = record['cpu_stats']['cpu_usage']['total_usage']
        pre_cpu = record['precpu_stats']['cpu_usage']['total_usage']

        cpu_delta = float(current_cpu - pre_cpu) if pre_cpu > 0 else 0

        current_total = record['cpu_stats']['system_cpu_usage']
        pre_total = record['precpu_stats'].get('system_cpu_usage', 0)
        total_delta = float(current_total - pre_total) if pre_total > 0 else 0

        cpu_used = None
        if cpu_delta > 0 and total_delta > 0:
            cpu_used = round(cpu_delta / total_delta * 100, 4)

        cores_used = None
        current_cores_used = record['cpu_stats']['cpu_usage']['percpu_usage']
        pre_cores_used = record['precpu_stats']['cpu_usage'].get('percpu_usage', [])

        cores_used = None
        if len(current_cores_used) > 0 and len(pre_cores_used) > 0:
            cores_used = len([
                i for i, (curr, prev) 
                in enumerate(zip(current_cores_used, pre_cores_used))
                if curr - prev > 0
            ])

        return {"cpu_used": cpu_used, "cores_used": cores_used}

    def get_memory_used() -> Dict:
        mem_cache_mb = record['memory_stats']['stats']['cache'] * 1e-6
        mem_total_used_mb = record['memory_stats']['usage'] * 1e-6
        total_memory = record['memory_stats']['limit'] * 1e-6
        return {
            "mem_cache_mb": mem_cache_mb,
            "mem_total_used_mb": mem_total_used_mb,
            "mem_used_mb": mem_total_used_mb - mem_cache_mb,
            "mem_used_%": mem_total_used_mb / total_memory
        }

    def get_pids() -> Dict:
        return {
            "pids": float(record['pids_stats']['current'])
        }

    metadata = {
        "name": record['name'],
        "time": record['time'],
        "timestamp": record['read']
    }

    for metric, value in get_block_io_mb_stats().items():
        yield {**metadata, "metric": metric, "value": value}

    for metric, value in get_cpu_stats().items():
        yield {**metadata, "metric": metric, "value": value}

    for metric, value in get_memory_used().items():
        yield {**metadata, "metric": metric, "value": value}

    for metric, value in get_pids().items():
        yield {**metadata, "metric": metric, "value": value}

def stream_docker_stats(container_names: List[str]) -> Generator[Dict, None, None]:
    processes = [
        mp.Process(target=start_container_collection, args=(container_name,))
        for container_name in container_names
    ]
    try:
        for process in processes:
            process.start()

        while True:
            record = data_queue.get()
            yield from process_record(record)
    except (KeyboardInterrupt, SystemExit) as err:
        ...
    except Exception as err:
        log.exception(err)
    finally:
        for process in processes:
            process.terminate()


def get_all_containers():
    return tuple(
        container.name
        for container in client.containers.list()
    )


if __name__ == '__main__':

    stats_stream = stream_docker_stats(get_all_containers())

    for record in stats_stream:
        print(record)



import pandas as pd
import numpy as np
import argparse
import math

from docker_load_graph import docker_helper as d
from docker_load_graph import plotting

def main():
    p = argparse.ArgumentParser()
    p.add_argument("--rows", type=int, default=3)
    p.add_argument("--cols", type=int, default=2)
    args = p.parse_args()

    max_rows = args.rows
    max_cols = args.cols

    ordered_valid_keys = [
        'block_io_mb_write',
        'block_io_mb_read',
        'mem_used_mb',
        'cpu_used',
        'cores_used',
        'pids',
        'mem_used_%',
        'mem_cache_mb',
        'mem_total_used_mb',
    ]

    df = pd.DataFrame()
    try:
        for record in d.stream_docker_stats(d.get_all_containers()):
            record['timestamp'] = pd.to_datetime(record['timestamp'])
            df = df.append(record, ignore_index=True)
            plotting.plot(
                df, 
                ordered_keys=ordered_valid_keys, 
                max_rows=max_rows,
                max_cols=max_cols
            )
    except KeyboardInterrupt as err:
        ...

if __name__ == '__main__':
    main()



from typing import List
import plotext as plt
import pandas as pd

def plot(df: pd.DataFrame, ordered_keys: List[str], max_rows: int, max_cols: int):
    trimmed_keys = ordered_keys[:max_rows * max_cols]

    plot_positions = [
        (row, col)
        for row in range(1, max_rows + 1)
        for col in range(1, max_cols + 1)
    ]

    key_to_position = dict(zip(trimmed_keys, plot_positions))

    plt.clp()

    plt.subplots(max_rows, max_cols)
    plt.colorless()

    plots_used = []

    for metric in trimmed_keys: 
        data = df[df.metric == metric]
        row, column = key_to_position[metric]
        plots_used.append((row, column))
        plt.subplot(row, column)
        plt.colorless()
        should_legend = bool(row == 1 and column == 1)
        plt.title(metric)
        max_y = 0
        min_y = None

        for name, group_by_name_with_nans in data.groupby('name'):
            group = group_by_name_with_nans.dropna()
            if len(group) > 0:
                max_y = max(group.value.max() * 2.0, max_y)
                min_y = min(group.value.min(), min_y) if min_y else group.value.min()
                plt.plot(
                    group.time.tolist(),
                    group.value.tolist(),
                    label=name if should_legend else None
                )

                plt.xticks(
                    group.time.tolist(),
                    [dt.time().strftime("%H:%M:%S") for dt in group.timestamp.tolist()]
                )
                plt.xlabel('time')

        if max_y > 0.0 and should_legend:
            plt.ylim(min_y, max_y)

    for row, column in set(plot_positions) - set(plots_used):
        plt.subplot(row, column)
        plt.colorless()

    plt.show()


